#include "globals.h"
#include "../dynarec/dynarec.h"
#include "regcache.h"

#define _eax_ 0
#define _ecx_ 1
#define _edx_ 2
#define _ebx_ 3
#define _esp_ 4
#define _ebp_ 5
#define _esi_ 6
#define _edi_ 7

#define ItIsARegisterNotToUse ((k==_esp_) || (k==_ebp_) || (k==_esi_)) //Don't use these registers
                                                                       //for mapping

/*****************************************************************/

x86regtyp x86reg[8]; /* Stores Mapped MIPS registers */


/******************************************************************
  Function: InitRegisterMap()
  Purpose:  This function will set each entry in the MIPS->x86 map 
            to -1 (unused)
  input:    nothing
  output:   nothing
*******************************************************************/
void InitRegisterMap()
{
    int k;

    for (k=0; k<8; k++)
    {
        x86reg[k].mips_reg  = -1;
        x86reg[k].x86reg    = k;
        x86reg[k].Is32bit   = 0;    // If 1, this register uses 32bit

        x86reg[k].IsDirty   = 0;    // If 1, this register will eventually need 
                                    // to be unmapped (written back to memory-- 
                                    // "flushed")
    }
}


/******************************************************************
  Function: MapRegister32()
  Purpose:  This function is called when you want to map a MIPS 
            General Purpose Register (GPR) to a native x86 register.
  input:    x86regtyp Conditions: The MIPS register to be mapped (with 
                                  its conditions--Is it Dirty? Is it
                                  32bit or 64bit?).
  output:   Conditions->x86reg  : Which x86 register it has been mapped to.
                                  Conditions->x86reg = -1: reg failed to map.
*******************************************************************/
uint8* MapRegister32(uint8* Dest, x86regtyp* Conditions)
{
    _int8 k;
    uint32 temp;

    /* If the MIPS register is already mapped, 
    find out which x86 register it is mapped to */

    for (k=0; k < 8; k++)
    {
        if (ItIsARegisterNotToUse);
        else if (x86reg[k].mips_reg == Conditions->mips_reg) // if mapped 
        {
            Conditions->x86reg = k;
            if (Conditions->IsDirty == 1)   x86reg[k].IsDirty = 1;
            if (Conditions->Is32bit == 1)   x86reg[k].Is32bit = 1;
            return(Dest);
        }
    }


    /* If the MIPS register is unmapped, map it to the next
    available x86 register */
    for (k=0; k < 8; k++)
    {
        if (ItIsARegisterNotToUse);
        else if (x86reg[k].mips_reg == -1) 
        {
            /* map it */
            x86reg[k].mips_reg  = Conditions->mips_reg;
            if (Conditions->IsDirty == 1)   x86reg[k].IsDirty = 1;
            if (Conditions->Is32bit == 1)   x86reg[k].Is32bit = 1;
            Conditions->x86reg  = k;

            temp = (uint32)&gHardwareState.GPR+(x86reg[k].mips_reg<<3);
            //if eax:
            //00416DC9 A1 A0 7A 4A 00       mov         eax,[_GPR+20h (004a7aa0)]
            if (k == 0)
            {
                PUTDST8KNOWN(Dest, 0xA1)
            }
            else
            //if ecx: (etc..)
            //00416DD0 8B 0D A4 7A 4A 00    mov         ecx,dword ptr [_GPR+24h (004a7aa4)]
            {
                uint8 temp2 = 5|(k<<3);
                PUTDST8KNOWN(Dest, 0x8B)    PUTDST8(Dest, temp2)
            }
            PUTDST32(Dest, temp)

            return(Dest);
        }
            
    }

    /* If there are no registers left to map, Success = -1 (failed) */
    Conditions->x86reg  = -1;
    return(Dest);
}


/******************************************************************/


uint8* WriteBackDirty(uint8* Dest, uint32 k)
{
    uint32 toGPR = (uint32)&gHardwareState.GPR[0];
    uint32 Target;
//if eax:
//00413EFE A3 80 83 49 00       mov         [_GPR],eax
//00413F03 C1 F8 1F             sar         eax,1Fh
//00413F06 A3 84 83 49 00       mov         [_GPR+4],eax
    if (k == 0)
    {
        Target = toGPR+(x86reg[0].mips_reg<<3);
        PUTDST8KNOWN(Dest, 0xA3)        PUTDST32(Dest, Target)
        PUTDST32KNOWN(Dest, 0xA31FF8C1) PUTDST32(Dest, (Target+4))
    }

//if ecx: (etc..)
//00413F0D 89 0D 80 83 49 00    mov         dword ptr [_GPR],ecx
//00413F13 C1 F9 1F             sar         ecx,1Fh
//00413F16 89 0D 84 83 49 00    mov         dword ptr [_GPR+4],ecx
    else
    {
        uint32 One = 0x5|(k<<3);
        uint32 Two = toGPR+(x86reg[k].mips_reg<<3);
        uint32 Three = 0xF8|k;

        PUTDST8KNOWN(Dest, 0x89)    PUTDST8(Dest, One)      PUTDST32(Dest, Two)
        PUTDST8KNOWN(Dest, 0xC1)    PUTDST8(Dest, Three)    PUTDST8KNOWN(Dest, 0x1F)
        PUTDST8KNOWN(Dest, 0x89)    PUTDST8(Dest, One)      PUTDST32(Dest, (Two+4))
    }

    return(Dest);
}




/******************************************************************
  Function: FlushAllRegisters32()
  Purpose:  This function is used when you want to "flush" all the 
            General Purpose Registers (GPR) that are mapped so that
            no registers are mapped. An unmapped array entry = -1.
  input:    uint8* Dest- Our recompiled code pointer (as usual)
  output:   uint8* Dest (as usual)
*******************************************************************/
uint8* FlushAllRegisters32(uint8* Dest)
{
    int k;

    for(k=0; k<8; k++)
    {
        if (ItIsARegisterNotToUse);
        else if (x86reg[k].mips_reg != -1)
        {
            if (x86reg[k].IsDirty == 1)
            {
                Dest = WriteBackDirty(Dest, k);
                x86reg[k].IsDirty      = 0;
            }
            x86reg[k].mips_reg = -1;
            x86reg[k].Is32bit   = 0;
        }
    }
    
    return(Dest);
}


/******************************************************************/


uint8* FlushAllBut2(uint8* Dest, _int8 The1stOneNotToFlush, _int8 The2ndOneNotToFlush)
{
    int k;
    for(k=0; k<8; k++)
    {
        if (ItIsARegisterNotToUse);
        else if ((x86reg[k].mips_reg != -1) && (x86reg[k].mips_reg != The1stOneNotToFlush) && (x86reg[k].mips_reg != The2ndOneNotToFlush))
        {
            if (x86reg[k].IsDirty == 1)
            {
                Dest = WriteBackDirty(Dest, k);
                x86reg[k].IsDirty      = 0;
            }
            x86reg[k].mips_reg = -1;
            x86reg[k].Is32bit   = 0;
        }
    }

    return(Dest);
}